/****************************************************************************\
|                                                                            |
|                               LS for MSDOS(tm)                             |
|                                                                            |
| This is an MSDOS version of the Unix(tm) directory lister "ls".            |
| The program was developed under MSDOS 3.1 and runs on versions 2.0 and up. |
| I am placing it without conditions in the public domain, it may be copied  |
| and distributed freely. I ask that you do not sell it, and that you leave  |
| this notice intact.                                                        |
| If you find any bugs, I'd appreciate a report.                             |
|                                                                            |
| Yaniv Hakohen                                                              |
|                                                                            |
| E-Mail:   yaniv@taurus.bitnet                                              |
|  or       tykucinski@nyuacf.bitnet                                         |
| MailMail: Computer Science Dept., Tel-Aviv University, Tel-Aviv, ISRAEL.   |
|  or       3 East, 66 St., New York, NY, 10021.                             |
|                                                                            |
\****************************************************************************/



/*
 * USAGE:
 *      ls [ -FRldaestrq1 ] [ filenames and/or directory names ]
 *      / can be used instead of -
 *
 * OPTIONS:
 *      -F executable file names are appended with '*' character.
 *         directory names are appended with '\' character.
 *      -R after normal listing, all entries which are directories
 *         are recursively listed.
 *      -l long format listing includes:
 *         permission bits, file sizes (logical and real), date.
 *         permission bits are
 *         [d]irectory [x]ecutable [r]ead_only [a]rchived [s]ystem [h]idden.
 *      -d if argument is directory, list only it's name. Use with -l
 *         to get status of directory.
 *      -a include hidden files in listing. The files "." and ".." are
 *         considered hidden and are revealed also with this option.
 *      -e sort by extension, and then alphabetically. (Default is alpha.)
 *      -s sort by size instead of default sort.
 *      -t sort by time instead of default sort.
 *      -r reverse the order of sort to get reverse alphabetic or oldest
 *         first or largest first, as appropriate.
 *      -q substitute unprintable characters within filenames with '?'.
 *      -1 list each entry on a separate line.
 *      -v prepends volume name and remaining diskspace to list.
 *
 * The environment variable LSOPTS is scanned for options every time
 * ls is invoked. If you always want ls to be -F and -e for instance,
 * add the line "set LSOPTS=-F-e" to your autoexec.bat file.
 *
 * NOTES:
 *      Compiled with Microsoft C Version 3, small memory model.
 *      Must be given larger stack than default. I used 8K and it was
 *      fine with a recursive list of 500 files which were up to 5 subdirs
 *      deep.
 *      The command "cl ls.c -link /STACK:8192" or some such, is fine.
 *      Do NOT link with the ssetargv.obj module!
 *
 * UNIX COMPATABILITY:
 *      This is NOT the standard Unix ls, but functionally simmilar.
 *      Output format is not totaly consistent with Unix.
 *      Some other notable differences:
 *      1)  Default sort is not alphabetical but first by extension,
 *          then by filename. You can have a normal alphasort with -e.
 *      2)  -l "permission bits" are dxrash i.e. directory, file is executable,
 *          file is read only, file is archived, system file, hidden file.
 *          "executable" means that the extension is EXE, COM or BIN.
 *      3)  -l has 2 size fields: logical size and real (disk space used) size.
 *      4)  -s option sorts by size, does not list in Kbytes.
 *      5)  Several Unix options are not implemented, some for objective
 *          reasons (-gLuciA), some I will do (maybe...) in a future
 *          release (-fC).
 *
 * BUGS:
 *      - References to the root directory with things like . and .. (and
 *      ..\foo, ..\..\bar and such) are not understood.
 *      - It's kind of enormous.
 */

#include <stdio.h>
#include <dos.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <sys\types.h>
#include <sys\stat.h>

#define NOEXIST         0
#define ISFILE          1
#define ISDIR           -1
#define FIRSTENT        0
#define NEXTENT         1
#define READONLY        01
#define HIDDEN          02
#define SYSTEM          04
#define VOLABEL         010
#define SUBDIR          020
#define ARCHIVE         040
#define EXECABLE        0100
#define MAXPATH         64

union   filedate {
        unsigned short  d;
        struct  {
                unsigned int mday:5;
                unsigned int month:4;
                unsigned int year:7;
        } b;
};

union   filetime {
        unsigned short  t;
        struct  {
                unsigned int twosec:5;
                unsigned int min:6;
                unsigned int hour:5;
        } b;
};

/*
 * A directory entry is:
 */
struct dirent {
        char                    fname[13];
        char                    attrib;
        long                    size;
        union   filetime        ftime;
        union   filedate        fdate;
        struct  dirent          *next;  /* for linked lists */
        char                    *orgpath;
        char                    orgdrive;
};
#define DIRENTSIZ       sizeof(struct dirent)
#define DIRENTPSIZ      sizeof(struct dirent *)

/* Disk Stats */
struct  disk {
        char    volname[12];
        char    drive;
        long    freespace;
        int     clusize;
};

/* Globals */
int     Rflag = 0, Fflag = 0, lflag = 0, aflag = 0, tflag = 0, sflag = 0,
        vflag = 0, dflag = 0, qflag = 0, eflag = 0, oneflag = 0,
        reverse_sort = 0;
int     nitems = 0, needcr = 0; /* for controlling crlf's on output */
int     alphasort(), timesort(), sizesort();
int     (*entry_compare)() = alphasort;
struct  disk dstat;
struct  dirent  *dralloc(), **allocdirparray();
char    *argv0, *shortname(), *fullpath(), curdrv();

main(argc, argv)
char    **argv;
{
        struct  dirent  *first = dralloc(), *last = first, *p, **sortarray;
        int     i, nerrs = 0, nfiles = 0, ndirs = 0, nentries, *spstat;
        char    cwd[MAXPATH + 1], *env = getenv("LSOPTS");

        argv0 = shortname(*argv);
        if (env != NULL)
                getflags(env);
        while (argc-- > 0 && (**++argv == '-' || **argv == '/')) {
                getflags(++*argv);
        }
        if (dflag)
                Rflag = 0;
        getcwd(cwd, MAXPATH);
        if (!argc) {
                if (Rflag || vflag)
                        dirheader(cwd);
                lsdir (cwd);
                exit(0);
        }
        if ((spstat = (int *) calloc(argc, sizeof (int))) == NULL)
                memerr();
        for (i = 0; i < argc; i++) {
                char    argdrive = argv[i][1] == ':' ?
                                        toupper(argv[i][0]) : curdrv();
                spstat[i] = NOEXIST;
                if (getdirent(FIRSTENT, argv[i], last, HIDDEN | SYSTEM | SUBDIR,
                        if (spstat[i] = exists(argv[i]))
                                continue;
                        fprintf(stderr, "%s: %s not found\n", argv0, argv[i]);
                        nerrs++;
                        continue;
                }
                last->orgdrive = argdrive;
                last->orgpath = argv[i];
                if (last->attrib & SUBDIR) ndirs++; else nfiles++;
                last->next = dralloc();
                last = last->next;
                while (getdirent(NEXTENT, 0, last, 0, 1) > 0) {
                        last->orgdrive = argdrive;
                        last->orgpath = argv[i];
                        if (last->attrib & SUBDIR) ndirs++; else nfiles++;
                        last->next = dralloc();
                        last = last->next;
                }
        }
        nentries = nfiles + ndirs;
        if (nerrs && nfiles)
                putchar('\n');

        sortarray = allocdirparray(nentries);
        for (p = first, i = 0; i < nentries; i++, p = p->next)
                sortarray[i] = p;
        qsort(sortarray, nentries, DIRENTPSIZ, entry_compare);

        for (i = 0; i < nentries; i++)
                if (!(sortarray[i]->attrib & SUBDIR) || dflag) {
                        char    *arg = sortarray[i]->orgpath;
                        printdirent(sortarray[i], sortarray[i]->orgdrive);
                }
        crlf();

        for (i = 0; !dflag && i < nentries; i++)
                if ((sortarray[i]->attrib & SUBDIR) && *(sortarray[i]->fname) !=
                        char    *newdir = fullpath(sortarray[i]->orgpath, sortar
                        dirheader(newdir);
                        lsdir (newdir);
                }
        for (i = 0; i < argc; i++)
                if (spstat[i]) {
                        char    olddrive = *cwd,
                                newdrive = argv[i][1] == ':' ? *argv[i] : *cwd,
                                oldplace[MAXPATH + 1];

                        if (newdrive != olddrive)
                                if (chdrive(newdrive) < 0) {
                                        fprintf(stderr, "\n%s: no '%c:' drive\n"
                                        continue;
                                }
                        getcwd(oldplace, MAXPATH);
                        chdir(argv[i]);
                        getcwd(cwd, MAXPATH);
                        dirheader(cwd);
                        lsdir(cwd);
                        chdir(oldplace);
                        if (newdrive != olddrive)
                                chdrive(olddrive);
                }
        exit(0);
}

lsdir (dirpath)
char    *dirpath;
{
        struct  dirent  *first, *p, *q, **f_entry;
        char    *temp, *mkkid();
        char    path[MAXPATH + 1];
        int     nentries = 1, i;

        strcpy(path, dirpath);
        if (*(temp = path + strlen(path) - 1) == ':' || *temp == '\\')
                strcpy(++temp, "*.*");
        else
                strcpy(++temp, "\\*.*");
        if (getdirent(FIRSTENT, path, p = first = dralloc(),
                                        HIDDEN | SYSTEM | SUBDIR, 1) < 0) {
                fprintf(stderr, "%s: %s not found\n", argv0, path);
                return;
        }
        else
                while (getdirent(NEXTENT, 0, q = dralloc(), 0, 1) > 0) {
                        nentries++;
                        p->next = q;
                        p = q;
                }
        f_entry = allocdirparray(nentries);
        for (p = first, i = 0; i < nentries; p = p->next, i++)
                f_entry[i] = p;
        qsort(f_entry, nentries, DIRENTPSIZ, entry_compare);
        for (i = 0; i < nentries; i++)
                printdirent(f_entry[i], *dirpath);
        crlf();
        if (Rflag) {
                for (i = 0; i < nentries; i++)
                        if ((f_entry[i]->attrib & SUBDIR) && *(f_entry[i]->fname
                                char    kid[MAXPATH + 1];
                                dirheader(mkkid(kid, dirpath, f_entry[i]->fname)
                                lsdir (kid);
                        }
        }
        for (i = 0; i < nentries; free(f_entry[i++]));
        free(f_entry);
}

char *
mkkid(kid, oldpath, addition)
char    *kid, *oldpath, *addition;
{
        char    *temp;
        strcpy(kid, oldpath);
        if (*(temp = kid + strlen(kid) - 1) != '\\')
                *++temp = '\\';
        strcpy(++temp, addition);
        return kid;
}

char *
catname(d)
struct dirent *d;
{
        static char fnamebuf[14];
        char    *p, *q;

        if (qflag)
                for (p = d->fname; *p; p++)
                        if (*p < ' ' || *p > 0x7e)
                                *p = '?';
        for (p = d->fname, q = fnamebuf; *q = tolower(*p++); q++);
        if (Fflag)
            *q++=d->attrib & SUBDIR ?'\\':(d->attrib & EXECABLE ?'*':'\0');
        *q = '\0';
        return fnamebuf;
}

ls_simple(d)
struct dirent *d;
{

        extern  needcr, nitems;
        char    *eol = oneflag || !(needcr = (++nitems % 5)) ? "\n" : "";

        printf ("%-15s%s", catname(d), eol);
}

ls_l(d)
struct dirent *d;
{
        static  char    *monthname[12] = {
                "Jan", "Feb", "Mar", "Apr", "May", "Jun",
                "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
        };
        long    realsize;

        realsize = d->size % dstat.clusize ?
                ((d->size / dstat.clusize) + 1) * dstat.clusize : d->size;
        printf("%c%c%c%c%c%c %6ld %6ld   %s %2d %4d  %02d:%02d   %s\n",
                d->attrib & SUBDIR   ? 'd' : '-',
                d->attrib & EXECABLE ? 'x' : '-',
                d->attrib & READONLY ? 'r' : '-',
                d->attrib & ARCHIVE  ? '-' : 'a',
                d->attrib & SYSTEM   ? 's' : '-',
                d->attrib & HIDDEN   ? 'h' : '-',
                d->size,
                realsize,
                monthname[d->fdate.b.month - 1],
                d->fdate.b.mday,
                d->fdate.b.year + 1980,
                d->ftime.b.hour,
                d->ftime.b.min,
                catname(d));
}

printdirent(d, drivename)
struct dirent *d;
char    drivename;
{
        if ((d->attrib & HIDDEN) && !aflag)
                return;
        if (lflag) {
                diskstat(drivename, &dstat);
                ls_l(d);
        }
        else
                ls_simple(d);
}

crlf()
{
        if (needcr) {
                putchar ('\n');
                needcr = nitems = 0;
        }
}

volinfo(drive)
char    drive;
{
        diskstat(drive, &dstat);
        if (*(dstat.volname))
                printf("==== Volume %s", dstat.volname);
        printf(" ==== %ld bytes free", dstat.freespace);
}

dirheader(s)
char    *s;
{
        printf("\n==== %s ", s);
        if (vflag)
                volinfo(*s);
        putchar('\n');
}

breakfname(wholename, fn, ex)
char    *wholename, *fn, *ex;
{
        if (*wholename == '.') {
                strcpy(fn, wholename);
                *ex = '\0';
                return;
        }
        while (*wholename && *wholename != '.')
                *fn++ = *wholename++;
        *fn = '\0';
        if (*wholename)
                wholename++;
        while (*ex++ = *wholename++);
}

alphasort(d1, d2)
struct dirent **d1, **d2;
{
        int     retval;
        if (!eflag)
                retval = strcmp ((*d1)->fname, (*d2)->fname);
        else {
                char    fn1[9], fn2[9], ex1[4], ex2[4];

                breakfname((*d1)->fname, fn1, ex1);
                breakfname((*d2)->fname, fn2, ex2);
                if (!(retval = strcmp(ex1, ex2)))   /* extensions equal? */
                        retval = strcmp(fn1, fn2);  /* then compare on name */
        }
        return reverse_sort ? -retval : retval;
}

sizesort(d1, d2)
struct dirent **d1, **d2;
{
        long    diff = (*d1)->size > (*d2)->size;
        int     retval = diff > 0 ? 1 : -1;
        return  reverse_sort ? -retval : retval;
}

timesort(d1, d2)
struct dirent **d1, **d2;
{
        int     retval;

        if ((*d1)->fdate.d > (*d2)->fdate.d)
                retval = 1;
        else if ((*d1)->fdate.d == (*d2)->fdate.d)
                retval = (*d1)->ftime.t > (*d2)->ftime.t ? 1 : -1;
        else
                retval = -1;
        return reverse_sort ? -retval : retval;
}

struct  dirent  *
dralloc()
{
        struct  dirent  *new;

        if ((new = (struct dirent *) malloc(DIRENTSIZ)) == NULL)
                memerr();
        new->next = NULL;
        return new;
}

struct  dirent  **
allocdirparray(npointers)
{
        struct  dirent **newarray;
        if ((newarray =
                (struct dirent **) calloc(npointers, DIRENTPSIZ)) == NULL)
                        memerr();
        return newarray;
}

memerr()
{
        fprintf(stderr, "%s: out of memory\n", argv0);
        exit(1);
}

char    *help[] = {
"USAGE:  ls [-FRldaestrq1] [filenames and/or directories]",
" or  :  ls -help",
"Note: A slash '/' can be used instead of a minus '-' before option letters.",
"Options:",
"       -help (or -h or -H or -? or /help or /h or /H or /?) - get this help.",
"       -F executable file names are appended with '*' character.",
"          directory names are appended with '\\' character.",
"       -R recursively list all sub-directories.",
"       -l long format listing includes: permission bits, file sizes ",
"          (logical and real), date. permission bits are:",
"          [d]irectory [x]ecutable [r]ead_only [a]rchived [s]ystem [h]idden.",
"       -d if argument is directory, list only it's name. Use with -l.",
"       -a include hidden files in listing.",
"       -e sort by extension, then alphabetically. (Default is alpha.)",
"       -s sort by size instead of default sort.",
"       -t sort by time instead of default sort.",
"       -r reverse the order of sort to get reverse alphabetic or oldest",
"          first or largest first, as appropriate.",
"       -q substitute unprintable characters within filenames with '?'.",
"       -1 list each entry on a separate line.",
"       -v volume name and free diskspace added to list.\n",
" ",
"LS V1.1 by Yaniv Hakohen is Public Domain, 1986.",
""
};

dohelp()
{
        char    **helpp = help;
        while (**helpp)
                puts(*helpp++);
        exit(0);
}

getdirent(which, pathname, d, filemask, calc_exe)
char    *pathname;
struct  dirent  *d;
unsigned char filemask;
{
        static  unsigned char   DTA[128];
        static  union   REGS    callregs, retregs;
        static  char            asczbuf[65];
        char                    *p =    asczbuf;
        long                    *l;
        union   filetime        *ft;
        union   filedate        *fd;

        if (which == FIRSTENT) {
                callregs.h.ah = 0x1a;
                callregs.x.dx = (unsigned int) DTA;
                intdos(&callregs, &retregs);
                while (*p = *pathname++)
                        p++;
                strcpy(p, "           $");
                callregs.h.ah = 0x4e;
                callregs.x.dx = (unsigned int) asczbuf;
                callregs.x.cx = filemask;
        }
        else
                callregs.h.ah = 0x4f;
        intdos(&callregs, &retregs);
        if (retregs.x.cflag)
                return -retregs.x.ax;
        strcpy(d->fname, (DTA + 30));
        if (*(d->fname) == '.')
                DTA[21] |= HIDDEN;
        d->attrib = DTA[21];
        p = d->fname + strlen(d->fname) - 4;
        if (calc_exe &&
            !(strcmp(p, ".BAT") && strcmp(p, ".EXE") && strcmp(p, ".COM")))
                d->attrib |= EXECABLE;
        ft = (union filetime *) (DTA + 22);
        fd = (union filedate *) (DTA + 24);
        l = (long *) (DTA + 26);
        d->size = *l;
        d->ftime = *ft;
        d->fdate = *fd;
        return 1;
}

char    *shortname(path)
char    *path;
{
        for (;*path = tolower(*path); path++);
        while (*--path != '.');
        *path = '\0';
        while (*--path != '\\');
        return ++path;
}

diskstat(driveletter, dsk)
char    driveletter;
struct  disk *dsk;
{
        union   REGS    callregs, retregs;
        long    ac, bs, sc, avail;
        struct  dirent  statent;
        static  char    volpath[7] = "?:\\*.*";
        static  char    lastdrive;

        if (lastdrive == driveletter)
                return;
        lastdrive = driveletter;
        dsk->drive = toupper(driveletter);
        if (!(dsk->drive)) {
                callregs.h.ah = 0x19;
                intdos(&callregs, &retregs);
                dsk->drive = retregs.h.al + 'A';
        }
        *volpath = dsk->drive;
        callregs.h.ah = 0x36;
        callregs.h.dl = dsk->drive - '@';
        intdos(&callregs, &retregs);
        ac = (long)retregs.x.bx;
        bs = (long)retregs.x.cx;
        if ((sc = (long)retregs.x.ax) == 0xffffL)
                return -1;
        dsk->clusize = bs * sc;
        dsk->freespace = ac * dsk->clusize;
        if (getdirent(FIRSTENT, volpath, &statent, VOLABEL, 0) < 0)
                return -1;
        {
                char *p = dsk->volname, *q = statent.fname;
                while (*p = *q++)
                        if (*p != '.')
                                p++;
        }
        return 0;
}

char    *dir_of_file(path)
char    *path;
{
        static  char    h[MAXPATH + 1];
        char    *p;

        strcpy(h, path);
        for (p = h; *p; p++)
                *p = toupper(*p);
        while (--p >= h && *p != '\\' && *p != ':');
        if (p < h)
                return ".";
        if (*p == '\\') {
                if (p == h)
                        return "\\";
                if (*(p - 1) == ':') {
                        *(p + 1) = '\0';
                        return h;
                }
                *p = '\0';
                return h;
        }
        return ".";
}

char    *fullpath(path, newtail)
char    *path, *newtail;
{
        char    orgcwd[MAXPATH + 1],    dir[MAXPATH + 1],
                orgdrv,                 drv;
        char    *retval, *p, *q;

        getcwd(orgcwd, MAXPATH);
        orgdrv = *orgcwd;
        drv = path[1] == ':' ? toupper(path[0]) : orgdrv;
        if (drv != orgdrv) {
                if (chdrive(drv) < 0)
                        return NULL;
                getcwd(orgcwd, MAXPATH);
        }
        if (chdir(dir_of_file(path)) < 0)
                return NULL;            /* illegal path */
        getcwd(dir, MAXPATH);
        chdir(orgcwd);
        if (drv != orgdrv)
                chdrive(orgdrv);
        if ((retval = malloc(strlen(dir) + strlen(newtail) + 2)) == NULL)
                memerr();
        for (p = retval, q = dir; *p = *q; p++, q++);
        if (*--q != '\\')
                *p++ = '\\';
        strcpy(p, newtail);
        return retval;
}

chdrive(driveletter)
char    driveletter;
{
        union   REGS    callregs, retregs;
        char    newdrive;

        driveletter = toupper(driveletter);
        if (!isalpha(driveletter))
                return -1;
        callregs.h.ah = 0xe;
        callregs.h.dl = driveletter - 'A';
        intdos(&callregs, &retregs);
        callregs.h.ah = 0x19;
        intdos(&callregs, &retregs);
        newdrive = retregs.h.al + 'A';
        return newdrive == driveletter ? 0 : -1;
}

char
curdrv()
{
        union   REGS    callregs, retregs;

        callregs.h.ah = 0x19;
        intdos(&callregs, &retregs);
        return retregs.h.al + 'A';
}

exists(path)
char    *path;
{
        struct  stat    st;

        if (isdrive(path))
                return ISDIR;
        if (stat(path, &st) < 0)
                return NOEXIST;
        if ((st.st_mode & S_IFMT) == S_IFDIR)
                return ISDIR;
        return ISFILE;
}

isdrive(path)
char    *path;
{
        switch(strlen(path)) {
        case 1:
                return *path == '\\' ? 1 : 0;
        case 2:
                return (isalpha(*path++) && *path == ':');
        case 3:
                return (isalpha(*path++) && *path++ == ':' && *path == '\\');
        default:
                return 0;
        }
}

isalpha(c) /* why doesn't ctype.h's isalpha work for me? */
char    c;
{
        return c >= 'A' && c <= 'Z' || c >= 'a' && c <= 'z';
}

getflags(s)
char    *s;
{
        while (*s) {
                switch(*s) {
                case    '-':
                case    ' ':
                case    '\t': break;
                case    'h':
                case    'H':
                case    '?': dohelp(); break;
                case    'R': Rflag = 1; break;
                case    'F': Fflag = 1; break;
                case    'l': lflag = 1; break;
                case    'a': aflag = 1; break;
                case    'q': qflag = 1; break;
                case    'r': reverse_sort = 1; break;
                case    '1': oneflag = 1; break;
                case    'd': dflag = 1; break;
                case    'e': eflag = 1; break;
                case    'v': vflag = 1; break;
                case    't':
                        tflag = 1;
                        entry_compare = timesort;
                        break;
                case    's':
                        sflag = 1;
                        entry_compare = sizesort;
                        break;
                default:
                        fprintf(stderr, "%s: unknown flag -%c\n",
                                argv0, *s);
                        break;
                }
                s++;
        }
}
